.( ftrac.fs) cr

\ Copyright (C) 2005,2006,2007,2008,2013 Marcos Cruz (programandala.net)
\ License: http://programandala.net/license

\ Character translation tool
\ Program written in Forth; developed with Forth 5mx and Gforth.

\ --------------------------------------------------------------
\ History

\ 2005-11-13
\ Start. First working version.

\ 2006-06-04
\ Bug fixed: 'trac' translated an empty string into an arbitrary one
\ character string. To fix it, '?do' has replaced 'do'.

\ 2006-06-23
\ '?do' does not work yet in Forth 5mx. Meanwhile, 'ftrac' uses an
\ 'if' instead.

\ 2006-09-25
\ '?do' used in 'ftrac'.
  
\ 2006-09-26
\ A bit faster version of 'ftrac', also with '?do'.
\ Second version of 'c>s', 33% faster.
\ Third version of 'c>s', almost 1% faster than the second one.

\ 2007-03-30
\ 'trac-array:' improved to activate the array just when created.

\ 2008-04-27
\ String length implemented.
\ This way the final translation can be longer than 255.

\ 2013-12-11
\ Source code style and stack notation updated; fixed some comments.

\ 2013-12-12
\ 'bounds' used instead of the idiom 'over + swap';
\ '?trac@' renamed to 'trac@?'.

\ --------------------------------------------------------------

marker ftrac

true constant long-translation?

\ String length management:
defer sl!  defer sl+  defer sl@
long-translation?
[if]    ' ! is sl!  ' @ is sl@  ' cell+ is sl+
[else]  ' c! is sl!  ' c@ is sl@  ' char+ is sl+
[then]
: +count  ( ca1 -- ca2 u )
  dup sl+ swap sl@
  ;
variable trac-array  \ address of the current translation table
: trac-array:  ( "<spaces>name" -- )
  \ Create a translation array.
  create  here 256 cells  dup allot
  over trac-array !  \ activate the just created array
  erase
  does>  ( -- ) ( dfa ) trac-array !  \ activate it
  ;
: trac!  ( c ca len -- )
  \ Update the current translation array
  \ to translate a char to a string.
  here >r s, align  cells trac-array @ + r> swap !
  ;  
: (trac@)  ( n -- ca | 0 )
  \ Fetch the content of an element of the current translation array.
  cells trac-array @ + @
  ;
: trac@?  ( c -- ca len true | false )
  \ Fetch the translation of a char
  \ according to the current translation array.
  (trac@) dup if  count true  then
  ;
false [if]
: c>s  ( c -- ca )
  \ Create a one char counted string.
  \ First version.
  1 pad c! pad 1+ c! pad
  ;
[then]
create str1  1 c, 0 c, align
str1 1+ constant (str1)
false [if]
: c>s  ( c -- ca )
  \ Create a one char counted string.
  \ Second version, 33% faster.
  (str1) c! str1
  ;
[then]
: c>s  ( c -- ca len )
  \ Create a one char counted string.
  \ Third version, almost 1% faster than the second one
  \ (considering the 'count' no more needed after calling).
   (str1) c! (str1) 1
   ;
: trac@  ( c -- ca len )
  \ Fetch the translation of a char according to the current translation array.
  [ false ] [if]
    dup (trac@) dup if  nip  else  drop c>s  then
  [else]  \ Alternative code
    dup trac@? if  rot drop  else  c>s  then
  [then]
  ;
variable translation  \ address of the translation
variable >translation  \ pointer to the next char in the translation
: translation+!  ( ca len -- )
  \ Add a string to the current translation.
  dup >r >translation @ swap cmove  \ move chars
  r> dup >translation +!  \ update char pointer
  translation @ dup sl@ rot + swap sl!  \ update string count
  ;
: translation0  ( -- )
  \ Init the translation.
  here
  dup translation !  \ address
  0 over sl!  \ nul lenght
  sl+ >translation !  \ init the pointer to the first char
  ;
: trac  ( ca1 len1 -- ca2 len2 )
  \ Translate a string using the current translation array.
  [ false ] [if]  
    \ Old version without ?do
    translation0  0 2dup =
    if    2drop
    else  do  dup i + c@ trac@ translation+!  loop
    then  drop  translation @ count
  [then]
  [ false ] [if]  
    translation0
    0 ?do  dup i + c@ trac@ translation+!  loop
    drop  translation @ count
  [then]
  \ Benchmark (building a whole website with fhp on Forth 5mx):
  \   3560 seconds with the previous version
  \   3447 seconds with this version (3% faster)
  translation0  bounds
  ?do  i c@ trac@ translation+!  loop
  translation @ +count
  ;
: .trac-array  ( -- )
  \ Print the content of the current translation array.
  255 0 do
    i trac@? if
      i dup 3 .r emit space type cr
      \ more 0= if  leave  then
    then
  loop
  ;

\ --------------------------------------------------------------
\ Benchmark

false [if]

: benchmark  ( loops -- )
  >r time&date
  r@ 0 do bl c>s count 2drop loop
  -seconds cr . ." seconds the old one"
  time&date
  r@ 0 do bl c>s1 2drop loop
  -seconds cr .  ." seconds the new one"
  r> drop
  ;
 
[then]

.( ftrac.fs compiled) cr
